| $Id: drawline.s,v 1.13 2006/08/31 23:56:51 chris Exp $

/*
 * Fast line-drawing routine
 * Chris Williams
 */

/*
 * This routine uses the Bresenham line algorithm.
 * 
 * Optimizations used in this routine:
 * In the main loops, all of the branches were likely to occur, so I inverted
 * them and branched them out of the main loop. This results in fewer clock
 * cycles. Of course, this is at the cost of a little size (which is usually
 * less important than speed in this application).
 * 
 * Also, following Samuel Stearley's suggestions, I combined each conditional
 * branch with the following DBcc, used 16-bit masks, and moved other code
 * around to make it faster (and even smaller in some cases).
 */

	.global	DrawLine

| void DrawLine(short x1, short y1, short x2, short y2, short mode);
DrawLine:
	movem.l	%d3-%d6,-(%sp)
	move	4+4*4+0(%sp),%d0	| x1
	move	4+4*4+2(%sp),%d1	| y1
	move	4+4*4+4(%sp),%d2	| x2
	move	4+4*4+6(%sp),%d3	| y2
	move.l	CURRENT_SCREEN,%a0
	move	CURRENT_INCY,%d6
	
	cmp	%d2,%d0			| if (x1 > x0)
	blt.b	0f
	exg	%d0,%d2			| x1 <> x2
	exg	%d1,%d3			| y1 <> y2
0:	
	| %d2 = dx
	| %d3 = dy
	| %d4 = sincy
	
	move	%d6,%d4
	
	| calculate dx and dy
	
	sub	%d0,%d2			| dx = x2 - x1
	
	sub	%d1,%d3			| dy = y2 - y1
	bge.b	0f			| if (dy < 0)
	neg	%d3			| dy = -dy
	neg	%d4			| sincy = -sincy
0:	
	| %a0 = saddr
	
	| calculate the screen address and pixel mask
	
	muls	%d6,%d1
	move	%d0,%d5
	asr	#3,%d5
	and	#~1,%d5
	add	%d5,%d1
	adda.w	%d1,%a0
	
	| %d5 = mask
	
	not	%d0			| mask = 1 << (7 - (x1 % 8))
	and	#15,%d0
	moveq	#0,%d5
	bset	%d0,%d5
	
	| get the line mode
	
	move	4+4*4+8(%sp),%d0
	beq.b	DrawLineWhite
	subq.w	#1,%d0
	bne	DrawLineXor
	
	| fall through
	

	| %d0 = scratch
	| %d1 = scratch
	| %d2 = dx
	| %d3 = dy
	| %d4 = sincy
	| %d5 = mask
	| %d6 = scratch
	| %d7 = scratch
	| %a0 = saddr
	

.macro	DrawLineX	op,step,nostep
	cmp	%d3,%d2			| if (dx >= dy)
	blt.b	5f
	
	| more horizontal
	
	| %d0 = i
	| %d1 = yerr
	
	move	%d2,%d0			| i = dx
	move	%d2,%d1			| yerr = dx
	
	add	%d2,%d2			| dx *= 2
	add	%d3,%d3			| dy *= 2
	
0:	\op.w	%d5,(%a0)		| *saddr ?= mask	| 12
	ror.w	#1,%d5			| mask >>>= 1		| 8
	b\step.b	2f	| unlikely				| 8/10
	sub	%d3,%d1			| yerr -= dy		| 4
	dbmi	%d0,0b			| while (i--)		| 10/14
	bge.b	9f	| unlikely				| 8/10
	
1:	add	%d2,%d1			| yerr += dx		| 4
	add	%d4,%a0			| saddr += sincy	| 4
	dbra	%d0,0b			| while (i--)		| 10/14
	bra.b	9f
	
2:	addq.l	#2,%a0			| ++saddr		| 8
	sub	%d3,%d1			| yerr -= dy		| 4
	dbmi	%d0,0b			| while (i--)		| 10/14
	bmi.b	1b	| likely				| 8/10
	bge.b	9f	| unlikely				| 8/10
	| 46.625p + 26.125s
	
5:	
	| more vertical
	
	| %d0 = i
	| %d1 = xerr
	
	move	%d3,%d0			| i = dy
	move	%d3,%d1			| xerr = dy
	
	add	%d2,%d2			| dx *= 2
	add	%d3,%d3			| dy *= 2
	
0:	\op.w	%d5,(%a0)		| *saddr ^= mask	| 12
	add	%d4,%a0			| saddr += sincy	| 4
	sub	%d2,%d1			| xerr -= dx		| 4
	dbmi	%d0,0b			| while (i--)		| 10/14
	bge.b	9f	| unlikely				| 8/10
	
	add	%d3,%d1			| xerr += dy		| 4
	ror.w	#1,%d5			| mask >>>= 1		| 8
	db\step	%d0,0b			| while (i--)		| 10/14
	addq.l	#2,%a0			| ++saddr		| 8
	db\nostep	%d0,0b			| while (i--)		| 10/14
	| 34p + 31.125s
	
9:	movem.l	(%sp)+,%d3-%d6
	rts
.endm

DrawLineBlack:
	DrawLineX or,cs,cc

DrawLineWhite:
	not.b	%d5
	DrawLineX and,cc,cs

DrawLineXor:
	DrawLineX eor,cs,cc

	.if	0
| This is just a test program
a=27
b=11

x1=120; y1=20
x2=x1+a; y2=y1+b
x3=x2+b; y3=y2+a
x4=x3-b; y4=y3+a
x5=x4-a; y5=y4+b
x6=x5-a; y6=y5-b
x7=x6-b; y7=y6-a
x8=x7+b; y8=y7-a

_main:
	move	#-1,-(%sp)	| mode
	
	move	#10000-1,%d6
0:	
	move	#y2,-(%sp)	| y2
	move	#x2,-(%sp)	| x2
	move	#y1,-(%sp)	| y1
	move	#x1,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y3,-(%sp)	| y2
	move	#x3,-(%sp)	| x2
	move	#y2,-(%sp)	| y1
	move	#x2,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y4,-(%sp)	| y2
	move	#x4,-(%sp)	| x2
	move	#y3,-(%sp)	| y1
	move	#x3,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y5,-(%sp)	| y2
	move	#x5,-(%sp)	| x2
	move	#y4,-(%sp)	| y1
	move	#x4,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y6,-(%sp)	| y2
	move	#x6,-(%sp)	| x2
	move	#y5,-(%sp)	| y1
	move	#x5,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y7,-(%sp)	| y2
	move	#x7,-(%sp)	| x2
	move	#y6,-(%sp)	| y1
	move	#x6,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y8,-(%sp)	| y2
	move	#x8,-(%sp)	| x2
	move	#y7,-(%sp)	| y1
	move	#x7,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	move	#y1,-(%sp)	| y2
	move	#x1,-(%sp)	| x2
	move	#y8,-(%sp)	| y1
	move	#x8,-(%sp)	| x1
	bsr	DrawLine
	adda	#2*4,%sp
	
	dbra	%d6,0b
	
	adda	#2,%sp
	rts
	
CURRENT_SCREEN:
	.dc.l	0x4c00
CURRENT_INCY:
	.dc.w	30
_ti89:
_ti92plus:
_nostub:
	.endif
